﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using ThinkGeo.MapSuite.Core;
using ThinkGeo.MapSuite.Routing;

namespace ThinkGeo.MapSuite.RoutingExamples
{
    public class RoutePathAndDirection
    {
        private readonly IEnumerable<PointShape> stopPoints;
        private readonly IEnumerable<string> avoidableFeatureIds;

        private PointShape endPoint;
        private PointShape startPoint;
        private RoutingEngine routingEngine;
        private FeatureSource barrierFeatureSource;

        public RoutePathAndDirection()
        { }

        public RoutePathAndDirection(RoutingEngine routingEngine, PointShape startPoint, PointShape endPoint)
            : this(routingEngine, startPoint, endPoint, null)
        {
        }

        public RoutePathAndDirection(RoutingEngine routingEngine, PointShape startPoint, PointShape endPoint, FeatureSource barrierFeatureSource)
            : this(routingEngine, startPoint, endPoint, barrierFeatureSource, new string[] { })
        {
        }

        public RoutePathAndDirection(RoutingEngine routingEngine, PointShape startPoint, PointShape endPoint, FeatureSource barrierFeatureSource, IEnumerable<string> avoidableFeatureIds)
            : this(routingEngine, startPoint, endPoint, barrierFeatureSource, avoidableFeatureIds, new PointShape[] { })
        {
        }

        public RoutePathAndDirection(RoutingEngine routingEngine, PointShape startPoint, PointShape endPoint, FeatureSource barrierFeatureSource, IEnumerable<PointShape> stopPoints)
            : this(routingEngine, startPoint, endPoint, barrierFeatureSource, new string[] { }, stopPoints)
        {
        }

        public RoutePathAndDirection(RoutingEngine routingEngine, PointShape startPoint, PointShape endPoint, FeatureSource barrierFeatureSource, IEnumerable<string> avoidableFeatureIds, IEnumerable<PointShape> stopPoints)
        {
            this.stopPoints = stopPoints;
            this.avoidableFeatureIds = avoidableFeatureIds;
            this.barrierFeatureSource = barrierFeatureSource;

            this.routingEngine = routingEngine;
            this.startPoint = startPoint;
            this.endPoint = endPoint;

            routingEngine.RoutingAlgorithm.FindingRoute += RoutingAlgorithm_FindingRoute;
        }

        public PointShape StartPoint
        {
            get { return startPoint; }
            set { startPoint = value; }
        }

        public PointShape EndPoint
        {
            get { return endPoint; }
            set { endPoint = value; }
        }

        public IEnumerable<PointShape> StopPoints
        {
            get { return stopPoints; }
        }

        public RoutingEngine RoutingEngine
        {
            get { return routingEngine; }
            set { routingEngine = value; }
        }

        public FeatureSource BarrierFeatureSource
        {
            get { return barrierFeatureSource; }
            set { barrierFeatureSource = value; }
        }

        public RoutingResult GetRoute()
        {
            if (stopPoints != null && !stopPoints.Any())
            {
                RoutingResult result = routingEngine.GetRoute(startPoint, endPoint);
                return result;
            }
            else
            {
                Collection<PointShape> points = new Collection<PointShape>();
                points.Add(startPoint);

                foreach (PointShape stopPoint in stopPoints)
                {
                    points.Add(stopPoint);
                }

                points.Add(endPoint);

                // Filter the invalid point.
                Collection<PointShape> availablePoint = FilterInvalidPoints(points);

                RoutingResult result = null;
                for (int i = 0; i < availablePoint.Count - 1; i++)
                {
                    RoutingResult tempResult = routingEngine.GetRoute(availablePoint[i], availablePoint[i + 1]);
                    result = CombineRoutingResult(result, tempResult);
                }

                return result;
            }
        }

        private Collection<PointShape> FilterInvalidPoints(IEnumerable<PointShape> points)
        {
            Collection<PointShape> result = new Collection<PointShape>();

            barrierFeatureSource.Open();
            Collection<Feature> barrierFeatures = barrierFeatureSource.GetAllFeatures(ReturningColumnsType.NoColumns);
            foreach (PointShape point in points)
            {
                bool inBarrier = barrierFeatures.Any(a => a.Contains(new Feature(point)));
                if (!inBarrier) result.Add(point);
            }

            return result;
        }

        private RoutingResult CombineRoutingResult(RoutingResult baseResult, RoutingResult additionalResult)
        {
            if (baseResult == null)
            {
                baseResult = new RoutingResult();
                baseResult.Distance = additionalResult.Distance;
                baseResult.Weight = additionalResult.Weight;
                baseResult.Route = additionalResult.Route;

                additionalResult.Features.ForEach(a => baseResult.Features.Add(a));
                additionalResult.OrderedStops.ForEach(a => baseResult.OrderedStops.Add(a));
                additionalResult.RouteSegments.ForEach(a => baseResult.RouteSegments.Add(a));
            }
            else
            {
                baseResult.Distance += additionalResult.Distance;
                baseResult.Weight += additionalResult.Weight;

                Collection<Vertex> vertexes = new Collection<Vertex>();
                baseResult.Route.Vertices.ForEach(vertexes.Add);
                additionalResult.Route.Vertices.ForEach(vertexes.Add);

                baseResult.Route = new LineShape(vertexes);
                additionalResult.Features.ForEach(a => baseResult.Features.Add(a));
                additionalResult.OrderedStops.ForEach(a => baseResult.OrderedStops.Add(a));
                additionalResult.RouteSegments.ForEach(a => baseResult.RouteSegments.Add(a));
            }

            return baseResult;
        }

        private void RoutingAlgorithm_FindingRoute(object sender, FindingRouteRoutingAlgorithmEventArgs e)
        {
            Collection<string> beContainedFeatureIds = new Collection<string>();
            barrierFeatureSource.Open();

            Collection<string> startPointAdjacentIds = e.RouteSegment.StartPointAdjacentIds;
            Collection<string> endPointAdjacentIds = e.RouteSegment.EndPointAdjacentIds;

            foreach (string id in startPointAdjacentIds.Where(id => avoidableFeatureIds.Contains(id)))
            {
                beContainedFeatureIds.Add(id);
            }

            foreach (string id in endPointAdjacentIds.Where(id => avoidableFeatureIds.Contains(id)))
            {
                beContainedFeatureIds.Add(id);
            }

            // Remove the ones that be contained in the avoidable area
            foreach (string id in beContainedFeatureIds)
            {
                if (e.RouteSegment.StartPointAdjacentIds.Contains(id))
                {
                    e.RouteSegment.StartPointAdjacentIds.Remove(id);
                }
                if (e.RouteSegment.EndPointAdjacentIds.Contains(id))
                {
                    e.RouteSegment.EndPointAdjacentIds.Remove(id);
                }
            }
        }
    }
}